import contextlib
import logging
import multiprocessing
import os
import socket
import torch.multiprocessing
import torch.utils.data


def recreate_dataloader_with_distributed_sampler(dataloader: torch.utils.data.DataLoader):
    if isinstance(dataloader.sampler, torch.utils.data.RandomSampler):
        shuffle = True
    elif isinstance(dataloader.sampler, torch.utils.data.SequentialSampler):
        shuffle = False
    else:
        raise RuntimeError(f"DDP Trainer doesn't support the sampler {dataloader.sampler}. Please make a pull request.")

    # Create a new sampler. If shuffle is True, the default random seed 0 will be used.
    sampler = torch.utils.data.distributed.DistributedSampler(dataloader.dataset, shuffle=shuffle)
    return recreate_dataloader_with_new_sampler(dataloader, sampler)


def recreate_dataloader_with_new_sampler(dataloader: torch.utils.data.DataLoader, sampler: torch.utils.data.Sampler):
    return torch.utils.data.DataLoader(dataloader.dataset, dataloader.batch_size, sampler=sampler, num_workers=dataloader.num_workers,
                                       collate_fn=dataloader.collate_fn, pin_memory=dataloader.pin_memory, drop_last=dataloader.drop_last, timeout=dataloader.timeout,
                                       worker_init_fn=dataloader.worker_init_fn, prefetch_factor=dataloader.prefetch_factor)


def _get_available_port():
    """Find an available port number on the localhost."""
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def _wrap_function(process_index, target_function, args, logger_info):
    logging.getLogger().setLevel(logger_info[0])
    logging.getLogger('irisml').setLevel(logger_info[1])
    for h in logger_info[2]:
        logging.getLogger().addHandler(h)

    target_function(process_index, *args)


def spawn_processes(target_function, args, nprocs):
    """Spawn subprocesses and wait until one of them ends.

    The subprocess inherits logging handlers from the current process.

    Returns:
        SpawnContext
    """
    # Keep using the logger if it can be serialized and deserialized in the child process.
    # The Irisml's official log handlers are serializable, but sometimes other platform such as pytest might add a new log handler.
    pickleable_log_handlers = [h for h in logging.getLogger().handlers if hasattr(h, '__getstate__')]
    logger_info = (logging.getLogger().getEffectiveLevel(), logging.getLogger('irisml').getEffectiveLevel(), pickleable_log_handlers)

    # There is a bug in file_descriptor sharing that results in "bad value(s) in fds_to_keep" exception.
    # https://github.com/pytorch/pytorch/issues/31571
    torch.multiprocessing.set_sharing_strategy('file_system')

    # Set environment variables if they are not defined. AML can set those variables if DistributedConfig was used.
    if not os.getenv('MASTER_ADDR'):
        os.environ['MASTER_ADDR'] = 'localhost'
    if not os.getenv('MASTER_PORT'):
        os.environ['MASTER_PORT'] = str(_get_available_port())

    # This method will block until all processes end or any process fails.
    return torch.multiprocessing.spawn(_wrap_function, args=(target_function, args, logger_info), nprocs=nprocs, join=False)


def spawn_and_wait(target_function, args, nprocs):
    context = spawn_processes(target_function, args, nprocs)
    sentinels = [p.sentinel for p in context.processes]

    # Wait until one of the processes is terminated.
    multiprocessing.connection.wait(sentinels)

    return context


def all_gather(tensor):
    if not torch.distributed.is_initialized():
        raise RuntimeError("Distributed training is not initialized.")

    world_size = torch.distributed.get_world_size()
    rank = torch.distributed.get_rank()
    tensors = [torch.zeros_like(tensor) for _ in range(world_size)]
    torch.distributed.all_gather(tensors, tensor)

    tensors[rank] = tensor  # Make sure the local tensor keeps the gradients.
    return torch.cat(tensors, dim=0)



import dataclasses
import io
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Export the given model as ONNX."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        input_size: int = 224

    @dataclasses.dataclass
    class Outputs:
        data: bytes = None

    def execute(self, inputs):
        model = torch.nn.Sequential(inputs.model, inputs.model.predictor)
        x = torch.randn(1, 3, self.config.input_size, self.config.input_size)
        with io.BytesIO() as bytes_io:
            torch.onnx.export(model, x, bytes_io)
            return self.Outputs(bytes(bytes_io.getbuffer()))



import unittest
import torch
from irisml.tasks.make_image_text_contrastive_model import Task


class TestMakeImageTextContrastiveModel(unittest.TestCase):
    def test_simple(self):
        image_model = torch.nn.Linear(3, 3)
        text_model = torch.nn.Linear(3, 3)

        inputs = Task.Inputs(image_model, text_model)
        outputs = Task(Task.Config()).execute(inputs)

        model = outputs.model

        image_feature, text_feature, logit_scale = model((torch.rand(2, 3), torch.rand(2, 3)))
        self.assertIsNotNone(image_feature)
        self.assertIsNotNone(text_feature)
        self.assertIsNotNone(logit_scale)

        loss = model.criterion((image_feature, text_feature, logit_scale), torch.tensor([0, 1]))
        self.assertIsNotNone(loss)



import torch.optim


def build_optimizer(name: str, parameters, base_lr, weight_decay=0, momentum=0):
    if name == 'sgd':
        return torch.optim.SGD(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    elif name == 'adam':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay)
    elif name == 'amsgrad':
        return torch.optim.Adam(parameters, lr=base_lr, weight_decay=weight_decay, amsgrad=True)
    elif name == 'rmsprop':
        return torch.optim.RMSprop(parameters, lr=base_lr, weight_decay=weight_decay, momentum=momentum)
    else:
        raise ValueError(f"Unsupported optimizer: {name}")


class OptimizerFactory:
    def __init__(self, name: str, base_lr, weight_decay=0, momentum=0):
        self._name = name
        self._base_lr = base_lr
        self._weight_decay = weight_decay
        self._momentum = momentum

    def __call__(self, parameters):
        return build_optimizer(self._name, parameters, self._base_lr, self._weight_decay, self._momentum)



import inspect
import typing
import torch.utils.data


class Dataset(torch.utils.data.Dataset):
    """Dataset wrapper to support both image-only transforms and image-and-targets transforms."""
    def __init__(self, dataset, transform):
        self._dataset = dataset
        self._transform = transform
        self._num_transform_args = self._get_num_args(transform)
        assert 1 <= self._num_transform_args <= 2

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        item = self._dataset[index]
        assert len(item) == 2
        if self._num_transform_args == 1:
            return self._transform(item[0]), item[1]
        elif self._num_transform_args == 2:
            return self._transform(*item)

    @staticmethod
    def _get_num_args(function):
        args = inspect.getfullargspec(function).args
        if args[0] == 'self':
            return len(args) - 1
        return len(args)


def collate(batch) -> typing.List:
    """Handle batching.

    Supported batch formats are:
        [(Tensor[C, H, W], int), ...]  # multiclass classification
        [(Tensor[C, H, W], [int, ...]), ...]  # multiclass or multilabel classification
        [(Tensor[C, H, W], Tensor[T]), ...]  # multiclass or multilabel classification
        [(Tensor[C, H, W], [[tag, x, y, x2, y2], ...]), ...]  # object detection
        [(Tensor[C, H, W], Tensor[T, 5]), ...]  # object detection
        [Tensor[C, H, W], ...]  # prediction
    """

    if isinstance(batch[0], tuple):
        return [_collate_tensor([b[0] for b in batch]), _collate_targets([b[1] for b in batch])]
    elif isinstance(batch[0], torch.Tensor):
        return [_collate_tensor(batch), None]

    raise TypeError(f"Unexpected type: {type(batch[0])}")


def _collate_tensor(batch: typing.List[torch.Tensor]):
    if all(b.shape == (1,) for b in batch):
        return torch.cat(batch)

    if all(batch[0].shape == b.shape for b in batch):
        return torch.stack(batch, 0)
    return batch


def _collate_targets(batch: typing.List):
    if isinstance(batch[0], torch.Tensor):
        return _collate_tensor(batch)

    if isinstance(batch[0], list):
        if any(b and isinstance(b[0], list) for b in batch):  # batch = [[[...], ...], ...]
            return [torch.tensor(b).reshape(-1, 5) for b in batch]
        elif any(b and isinstance(b[0], int) for b in batch):  # batch = [[int, ...], ...]
            return _collate_tensor([torch.tensor(b) for b in batch])
        elif not any(batch):
            return batch
    elif isinstance(batch[0], int):
        return torch.tensor(batch)


def build_dataloader(dataset, transform, batch_size: int, shuffle=True, drop_last=True):
    my_dataset = Dataset(dataset, transform)
    return torch.utils.data.DataLoader(my_dataset, batch_size=batch_size, num_workers=4, shuffle=shuffle, drop_last=drop_last, pin_memory=True, collate_fn=collate)



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate accuracy of the given prediction results.

    This task supports only multiclass classification.
    The dataset format is (<image(torch.Tensor)>, <target(int)>).
    The prediction results format is torch.Tensor[(N, num_classes)].
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: torch.Tensor
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        accuracy: float = 0

    def execute(self, inputs):
        if len(inputs.predictions.shape) != 2:
            raise RuntimeError(f"Invalid input shape: {inputs.predictions.shape}")

        num_samples = inputs.predictions.size(0)
        _, predicted_max_indexes = inputs.predictions.max(dim=1)
        targets = torch.Tensor([b[1] for b in inputs.dataset])

        if len(targets) != num_samples:
            raise RuntimeError(f"The number of prediction results and the number of targets doesn't match. {len(targets)} vs {num_samples}")

        accuracy = ((predicted_max_indexes == targets).sum() / num_samples).item()

        logger.info(f"Accuracy: {accuracy:.3f}")
        return self.Outputs(accuracy)



import logging
import time
import torch.cuda
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def on_train_epoch_start(self, trainer, model, epoch_index):
        self._epoch_sum_loss = 0
        self._epoch_num_samples = 0
        self._num_iters = 0
        self._epoch_start = time.time()

    def on_train_epoch_end(self, trainer, model, epoch_index):
        loss = self._epoch_sum_loss / self._epoch_num_samples
        training_time = time.time() - self._epoch_start
        log = f"Epoch {epoch_index}: Training loss: {loss:.4f}, time: {training_time:.2f}s, iteration: {self._num_iters}"
        if trainer.device.type == 'cuda':
            total_memory = torch.cuda.get_device_properties(trainer.device).total_memory // (1024 * 1024)
            max_allocated = torch.cuda.max_memory_allocated(trainer.device) // (1024 * 1024)
            torch.cuda.reset_peak_memory_stats(trainer.device)
            log += f", CUDA max memory: {max_allocated}/{total_memory} MB"
        logger.info(log)

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        self._epoch_sum_loss += float(loss.item()) * len(batch[0])
        self._epoch_num_samples += len(batch[0])
        self._num_iters += 1
        return loss

    def on_validation_start(self, trainer, model):
        self._validation_sum_loss = 0
        self._validation_num_samples = 0
        self._validation_num_iters = 0

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        self._validation_sum_loss += float(loss.item()) * len(batch[0])
        self._validation_num_samples += len(batch[0])
        self._validation_num_iters += 1

    def on_validation_end(self, trainer, model):
        loss = self._validation_sum_loss / self._validation_num_samples
        logger.info(f"Validation loss: {loss:.4f}, iteration: {self._validation_num_iters}")



import unittest
from irisml.tasks.make_image_text_transform import Task


class TestMakeImageTextTransform(unittest.TestCase):
    def test_simple(self):
        def image_transform(x, y):
            return x, y

        def prompt_generator(x):
            return [f'A {x} A', f'B {x} B']

        def tokenizer(x):
            return f'token {x}'

        class_names = ['class_a', 'class_b']
        inputs = Task.Inputs(image_transform=image_transform, class_names=class_names, prompt_generator=prompt_generator, tokenizer=tokenizer)
        outputs = Task(Task.Config()).execute(inputs)

        self.assertIn(outputs.transform('image', 0), ((('image', 'token A class_a A'), 0), (('image', 'token B class_a B'), 0)))
        self.assertIn(outputs.transform('image', 0), ((('image', 'token A class_a A'), 0), (('image', 'token B class_a B'), 0)))
        self.assertIn(outputs.transform('image', 0), ((('image', 'token A class_a A'), 0), (('image', 'token B class_a B'), 0)))
        self.assertIn(outputs.transform('image', 1), ((('image', 'token A class_b A'), 1), (('image', 'token B class_b B'), 1)))
        self.assertIn(outputs.transform('image', 1), ((('image', 'token A class_b A'), 1), (('image', 'token B class_b B'), 1)))
        self.assertIn(outputs.transform('image', 1), ((('image', 'token A class_b A'), 1), (('image', 'token B class_b B'), 1)))



import dataclasses
import torch


@dataclasses.dataclass
class LrSchedulerConfig:
    name: str = 'cosine_annealing'


def build_lr_scheduler(config, optimizer, num_epochs):
    config = config or LrSchedulerConfig()

    if config.name == 'cosine_annealing':
        return torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, num_epochs)
    elif config.name == 'linear_decreasing':
        return torch.optim.lr_scheduler.LinearLR(optimizer, start_factor=1, end_factor=0, total_iters=num_epochs)

    raise ValueError(f"Unsupported lr scheduler name: {config.name}")


class LrSchedulerFactory:
    def __init__(self, config: LrSchedulerConfig, num_epochs: int):
        self._config = config
        self._num_epochs = num_epochs

    def __call__(self, optimizer):
        return build_lr_scheduler(self._config, optimizer, self._num_epochs)



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Create a prompt generator for a classification task.

    Supported generator:
        'clip_imagenet_short': Generates prompts that are introduced in the openai/CLIP repository.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        name: str = 'clip_imagenet_short'

    @dataclasses.dataclass
    class Outputs:
        generator: typing.Callable[[str], typing.List[str]]

    def execute(self, inputs):
        if self.config.name == 'clip_imagenet_short':
            gen = clip_imagenet_short_generator
        else:
            raise RuntimeError
        return self.Outputs(gen)

    def dry_run(self, inputs):
        return self.execute(inputs)


def clip_imagenet_short_generator(label_name: str) -> typing.List[str]:
    templates = ['itap of a {}.',
                 'a bad photo of the {}.',
                 'a origami {}.',
                 'a photo of the large {}.',
                 'a {} in a video game.',
                 'art of the {}.',
                 'a photo of the small {}.']
    return [t.format(label_name) for t in templates]



import copy
import dataclasses
import importlib
import logging
import typing
import torch
import irisml.core

from .build_dataloader import build_dataloader
from .build_lr_scheduler import LrSchedulerFactory, LrSchedulerConfig
from .build_optimizer import OptimizerFactory
from .ddp_trainer import DDPTrainer
from .trainer import Trainer

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Train a pytorch model.

    This is a simple training baseline.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        num_epochs: int
        batch_size: int = 1
        base_lr: float = 1e-5
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        lr_scheduler: typing.Optional[LrSchedulerConfig] = None
        momentum: float = 0
        optimizer: str = 'sgd'
        val_check_interval: typing.Optional[typing.Union[int, float]] = None
        weight_decay: float = 0
        num_processes: int = 1
        plugins: typing.List[str] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        train_dataset: torch.utils.data.Dataset
        train_transform: typing.Callable
        val_dataset: torch.utils.data.Dataset = None
        val_transform: typing.Callable = None

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None

    def execute(self, inputs):
        model = self.train(inputs)
        return self.Outputs(model)

    def dry_run(self, inputs):
        model = self.train(inputs, dry_run=True)
        return self.Outputs(model)

    def train(self, inputs, dry_run=False) -> torch.nn.Module:
        plugins = [self._load_plugin(p) for p in self.config.plugins]
        if dry_run:
            train_dataloader = val_dataloader = None
        else:
            train_dataloader = build_dataloader(inputs.train_dataset, inputs.train_transform, batch_size=self.config.batch_size)
            val_dataloader = inputs.val_dataset and build_dataloader(inputs.val_dataset, inputs.val_transform, batch_size=self.config.batch_size)
        device = self._get_device()
        model = copy.deepcopy(inputs.model)

        optimizer_factory = OptimizerFactory(self.config.optimizer, self.config.base_lr, self.config.weight_decay, self.config.momentum)
        lr_scheduler_factory = LrSchedulerFactory(self.config.lr_scheduler, self.config.num_epochs)

        trainer_parameters = {'model': model, 'lr_scheduler_factory': lr_scheduler_factory, 'optimizer_factory': optimizer_factory, 'plugins': plugins,
                              'device': device, 'val_check_interval': self.config.val_check_interval}

        if self.config.num_processes == 1:
            trainer = Trainer(**trainer_parameters)
        else:
            trainer = DDPTrainer(num_processes=self.config.num_processes, **trainer_parameters)

        if not dry_run:
            trainer.train(train_dataloader, val_dataloader, num_epochs=self.config.num_epochs)

        return trainer.model

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)

    def _load_plugin(self, plugin_name):
        try:
            plugin_module = importlib.import_module('irisml.tasks.train.plugins.' + plugin_name)
        except ModuleNotFoundError as e:
            raise RuntimeError(f"Plugin {plugin_name} is not found.") from e

        plugin_class = getattr(plugin_module, 'Plugin', None)
        return plugin_class()



import unittest
import torch
from irisml.tasks.evaluate_detection_average_precision import Task


class TestEvaluateDetectionAveragePrecision(unittest.TestCase):
    def test_evaluate(self):
        task = Task(Task.Config([0.5, 0.75, 0.9]))

        inputs = Task.Inputs(predictions=[torch.Tensor([[0, 0.5, 0, 0, 1, 1]])], dataset=[(None, torch.tensor([[0, 0, 0, 1, 1]]))])
        results = task.execute(inputs)
        self.assertEqual(results.results, [1.0, 1.0, 1.0])

        inputs = Task.Inputs(predictions=[torch.Tensor([[0, 0.5, 0.9, 0.9, 1, 1]])], dataset=[(None, torch.tensor([[0, 0, 0, 0.1, 0.1]]))])
        results = task.execute(inputs)
        self.assertEqual(results.results, [0.0, 0.0, 0.0])



from .plugin_base import PluginBase


class Plugin(PluginBase):
    def on_train_epoch_start(self, trainer, model, epoch_index):
        # Reset the random seed for DistributedSampler so that it can sample differently at each epoch.
        trainer.train_dataloader.sampler.set_epoch(epoch_index)



import unittest
import torch
from irisml.tasks.build_zero_shot_classifier import Task


class TestBuildZeroShotClassifier(unittest.TestCase):
    def test_simple(self):
        zero_feature = torch.ones(32)
        zero_feature[0] = 10
        one_feature = torch.ones(32)
        one_feature[1] = 10

        text_features = [one_feature, zero_feature, one_feature, zero_feature]
        text_classes = [1, 0, 1, 0]
        inputs = Task.Inputs(text_features, text_classes)
        outputs = Task(Task.Config(num_classes=2)).execute(inputs)

        self.assertEqual(torch.argmax(outputs.classifier(one_feature)), 1)
        self.assertEqual(torch.argmax(outputs.classifier(zero_feature)), 0)



import collections
import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Create a zero-shot classification layer."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        text_features: typing.List[torch.Tensor]
        text_classes: typing.List[int]

    @dataclasses.dataclass
    class Config:
        num_classes: int

    @dataclasses.dataclass
    class Outputs:
        classifier: torch.nn.Module = None

    def execute(self, inputs):
        num_examples = len(inputs.text_features)
        if len(inputs.text_classes) != num_examples:
            raise RuntimeError(f"The number of examples doesn't match. features={num_examples}, classes={len(inputs.text_classes)}")

        if max(inputs.text_classes) >= self.config.num_classes:
            raise RuntimeError(f"The max class index is higher than {self.config.num_classes}: {max(inputs.text_classes)}")
        feature_shape = inputs.text_features[0].shape
        logger.debug(f"Feature shape is {feature_shape}")

        features_per_class = collections.defaultdict(list)
        for c, f in zip(inputs.text_classes, inputs.text_features):
            features_per_class[c].append(f)

        embeddings_per_class = {}
        for c in features_per_class:
            class_embeddings = torch.stack(features_per_class[c])
            class_embeddings /= class_embeddings.norm(dim=-1, keepdim=True)
            class_embedding = class_embeddings.mean(dim=0)
            class_embedding /= class_embedding.norm()
            embeddings_per_class[c] = class_embedding
        weights_list = [embeddings_per_class[i] if i in embeddings_per_class else torch.zeros(*feature_shape) for i in range(self.config.num_classes)]
        weights = torch.stack(weights_list, dim=1).transpose(0, 1)  # TODO: Multiply with logit_scale?

        with torch.no_grad():
            classifier = torch.nn.Linear(weights.shape[1], weights.shape[0], bias=False)
            classifier.weight.copy_(weights)
        return self.Outputs(classifier)



import dataclasses
import random
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a transform function for image-text classification.

    The transform function accepts (image (PIL.Image), targets (int)).

    Notes:
        This transform function depends on python's global random generator.

    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: typing.Callable
        class_names: typing.List[str]
        prompt_generator: typing.Callable[[str], typing.List[str]]
        tokenizer: typing.Callable[[str], torch.Tensor]

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable

    def execute(self, inputs):
        transform = ImageTextClassificationTransform(inputs.image_transform, inputs.class_names, inputs.prompt_generator, inputs.tokenizer)
        return self.Outputs(transform)

    def dry_run(self, inputs):
        return self.execute(inputs)


class ImageTextClassificationTransform(torch.nn.Module):
    def __init__(self, image_transform, class_names, prompt_generator, tokenizer):
        super().__init__()
        self._image_transform = image_transform
        self._class_names = class_names
        self._prompt_generator = prompt_generator
        self._tokenizer = tokenizer

    def forward(self, inputs, targets):
        image = inputs
        image, targets = self._image_transform(image, targets)
        prompts = self._prompt_generator(self._class_names[targets])
        text = self._tokenizer(random.choice(prompts))
        return (image, text), targets



import dataclasses
import logging
import typing
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Create a classification prompt dataset."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        class_names: typing.List[str]
        prompt_generator: typing.Callable[[str], typing.List[str]]

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset = None

    def execute(self, inputs):
        text_lists = [(t, i) for i, label in enumerate(inputs.class_names) for t in inputs.prompt_generator(label)]
        logger.debug(f"Created a text dataset. The number of examples: {len(text_lists)}. The number of classes: {len(inputs.class_names)}")
        return self.Outputs(TextListDataset(text_lists))

    def dry_run(self, inputs):
        return self.execute(inputs)


class TextListDataset(torch.utils.data.Dataset):
    def __init__(self, text_list: typing.List[typing.Tuple[str, int]]):
        self._data = text_list

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]
